import unittest

import os

import dbt.flags
import dbt.compilation
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.parsed import NodeConfig, DependsOn
from dbt.contracts.graph.compiled import CompiledModelNode, InjectedCTE
from dbt.node_types import NodeType

from datetime import datetime


class CompilerTest(unittest.TestCase):
    def assertEqualIgnoreWhitespace(self, a, b):
        self.assertEqual(
            "".join(a.split()),
            "".join(b.split()))

    def setUp(self):
        dbt.flags.STRICT_MODE = True

        self.maxDiff = None

        self.root_project_config = {
            'name': 'root_project',
            'version': '0.1',
            'profile': 'test',
            'project-root': os.path.abspath('.'),
        }

        self.snowplow_project_config = {
            'name': 'snowplow',
            'version': '0.1',
            'project-root': os.path.abspath('./dbt_modules/snowplow'),
        }

        self.model_config = NodeConfig.from_dict({
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'vars': {},
            'quoting': {},
            'column_types': {},
            'tags': [],
        })

    def test__prepend_ctes__already_has_cte(self):
        ephemeral_config = self.model_config.replace(materialized='ephemeral')

        input_graph = Manifest(
            macros={},
            nodes={
                'model.root.view': CompiledModelNode(
                    name='view',
                    database='dbt',
                    schema='analytics',
                    alias='view',
                    resource_type=NodeType.Model,
                    unique_id='model.root.view',
                    fqn=['root_project', 'view'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(nodes=['model.root.ephemeral']),
                    config=self.model_config,
                    tags=[],
                    path='view.sql',
                    original_file_path='view.sql',
                    raw_sql='select * from {{ref("ephemeral")}}',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[InjectedCTE(id='model.root.ephemeral', sql='select * from source_table')],
                    injected_sql='',
                    compiled_sql=(
                        'with cte as (select * from something_else) '
                        'select * from __dbt__CTE__ephemeral')
                ),
                'model.root.ephemeral': CompiledModelNode(
                    name='ephemeral',
                    database='dbt',
                    schema='analytics',
                    alias='view',
                    resource_type=NodeType.Model,
                    unique_id='model.root.ephemeral',
                    fqn=['root_project', 'ephemeral'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(),
                    config=ephemeral_config,
                    tags=[],
                    path='ephemeral.sql',
                    original_file_path='ephemeral.sql',
                    raw_sql='select * from source_table',
                    compiled=True,
                    compiled_sql='select * from source_table',
                    extra_ctes_injected=False,
                    extra_ctes=[],
                    injected_sql=''
                ),
            },
            docs={},
            # '2018-02-14T09:15:13Z'
            generated_at=datetime(2018, 2, 14, 9, 15, 13),
            disabled=[],
            files={},
        )

        result, output_graph = dbt.compilation.prepend_ctes(
            input_graph.nodes['model.root.view'],
            input_graph)

        self.assertEqual(result, output_graph.nodes['model.root.view'])
        self.assertEqual(result.extra_ctes_injected, True)
        self.assertEqualIgnoreWhitespace(
            result.injected_sql,
            ('with __dbt__CTE__ephemeral as ('
             'select * from source_table'
             '), cte as (select * from something_else) '
             'select * from __dbt__CTE__ephemeral'))

        self.assertEqual(
            input_graph.nodes['model.root.ephemeral'].extra_ctes_injected,
            True)

    def test__prepend_ctes__no_ctes(self):
        input_graph = Manifest(
            macros={},
            nodes={
                'model.root.view': CompiledModelNode(
                    name='view',
                    database='dbt',
                    schema='analytics',
                    alias='view',
                    resource_type=NodeType.Model,
                    unique_id='model.root.view',
                    fqn=['root_project', 'view'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(),
                    config=self.model_config,
                    tags=[],
                    path='view.sql',
                    original_file_path='view.sql',
                    raw_sql=('with cte as (select * from something_else) '
                             'select * from source_table'),
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[],
                    injected_sql='',
                    compiled_sql=('with cte as (select * from something_else) '
                                  'select * from source_table')
                ),
                'model.root.view_no_cte': CompiledModelNode(
                    name='view_no_cte',
                    database='dbt',
                    schema='analytics',
                    alias='view_no_cte',
                    resource_type=NodeType.Model,
                    unique_id='model.root.view_no_cte',
                    fqn=['root_project', 'view_no_cte'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(),
                    config=self.model_config,
                    tags=[],
                    path='view.sql',
                    original_file_path='view.sql',
                    raw_sql='select * from source_table',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[],
                    injected_sql='',
                    compiled_sql=('select * from source_table')
                ),
            },
            docs={},
            generated_at='2018-02-14T09:15:13Z',
            disabled=[],
            files={},
        )

        result, output_graph = dbt.compilation.prepend_ctes(
            input_graph.nodes.get('model.root.view'),
            input_graph)

        self.assertEqual(
            result,
            output_graph.nodes.get('model.root.view'))
        self.assertTrue(result.extra_ctes_injected)
        self.assertEqualIgnoreWhitespace(
            result.injected_sql,
            output_graph.nodes.get('model.root.view').compiled_sql)

        result, output_graph = dbt.compilation.prepend_ctes(
            input_graph.nodes.get('model.root.view_no_cte'),
            input_graph)

        self.assertEqual(
            result,
            output_graph.nodes.get('model.root.view_no_cte'))
        self.assertTrue(result.extra_ctes_injected)
        self.assertEqualIgnoreWhitespace(
            result.injected_sql,
            output_graph.nodes.get('model.root.view_no_cte').compiled_sql)

    def test__prepend_ctes(self):
        ephemeral_config = self.model_config.replace(materialized='ephemeral')

        input_graph = Manifest(
            macros={},
            nodes={
                'model.root.view': CompiledModelNode(
                    name='view',
                    database='dbt',
                    schema='analytics',
                    alias='view',
                    resource_type=NodeType.Model,
                    unique_id='model.root.view',
                    fqn=['root_project', 'view'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(nodes=['model.root.ephemeral']),
                    config=self.model_config,
                    tags=[],
                    path='view.sql',
                    original_file_path='view.sql',
                    raw_sql='select * from {{ref("ephemeral")}}',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[InjectedCTE(id='model.root.ephemeral', sql='select * from source_table')],
                    injected_sql='',
                    compiled_sql='select * from __dbt__CTE__ephemeral'
                ),
                'model.root.ephemeral': CompiledModelNode(
                    name='ephemeral',
                    database='dbt',
                    schema='analytics',
                    alias='ephemeral',
                    resource_type=NodeType.Model,
                    unique_id='model.root.ephemeral',
                    fqn=['root_project', 'ephemeral'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(),
                    config=ephemeral_config,
                    tags=[],
                    path='ephemeral.sql',
                    original_file_path='ephemeral.sql',
                    raw_sql='select * from source_table',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[],
                    injected_sql='',
                    compiled_sql='select * from source_table'
                ),
            },
            docs={},
            generated_at='2018-02-14T09:15:13Z',
            disabled=[],
            files={},
        )

        result, output_graph = dbt.compilation.prepend_ctes(
            input_graph.nodes.get('model.root.view'),
            input_graph)

        self.assertEqual(result,
                         output_graph.nodes.get('model.root.view'))

        self.assertTrue(result.extra_ctes_injected)
        self.assertEqualIgnoreWhitespace(
            result.injected_sql,
            ('with __dbt__CTE__ephemeral as ('
             'select * from source_table'
             ') '
             'select * from __dbt__CTE__ephemeral'))

        self.assertTrue(output_graph.nodes['model.root.ephemeral'].extra_ctes_injected)

    def test__prepend_ctes__multiple_levels(self):
        ephemeral_config = self.model_config.replace(materialized='ephemeral')

        input_graph = Manifest(
            macros={},
            nodes={
                'model.root.view': CompiledModelNode(
                    name='view',
                    database='dbt',
                    schema='analytics',
                    alias='view',
                    resource_type=NodeType.Model,
                    unique_id='model.root.view',
                    fqn=['root_project', 'view'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(nodes=['model.root.ephemeral']),
                    config=self.model_config,
                    tags=[],
                    path='view.sql',
                    original_file_path='view.sql',
                    raw_sql='select * from {{ref("ephemeral")}}',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[InjectedCTE(id='model.root.ephemeral', sql='select * from source_table')],
                    injected_sql='',
                    compiled_sql='select * from __dbt__CTE__ephemeral'
                ),
                'model.root.ephemeral': CompiledModelNode(
                    name='ephemeral',
                    database='dbt',
                    schema='analytics',
                    alias='ephemeral',
                    resource_type=NodeType.Model,
                    unique_id='model.root.ephemeral',
                    fqn=['root_project', 'ephemeral'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(),
                    config=ephemeral_config,
                    tags=[],
                    path='ephemeral.sql',
                    original_file_path='ephemeral.sql',
                    raw_sql='select * from {{ref("ephemeral_level_two")}}',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[InjectedCTE(id='model.root.ephemeral_level_two', sql='select * from source_table')],
                    injected_sql='',
                    compiled_sql='select * from __dbt__CTE__ephemeral_level_two' # noqa
                ),
                'model.root.ephemeral_level_two': CompiledModelNode(
                    name='ephemeral_level_two',
                    database='dbt',
                    schema='analytics',
                    alias='ephemeral_level_two',
                    resource_type=NodeType.Model,
                    unique_id='model.root.ephemeral_level_two',
                    fqn=['root_project', 'ephemeral_level_two'],
                    package_name='root',
                    root_path='/usr/src/app',
                    refs=[],
                    sources=[],
                    depends_on=DependsOn(),
                    config=ephemeral_config,
                    tags=[],
                    path='ephemeral_level_two.sql',
                    original_file_path='ephemeral_level_two.sql',
                    raw_sql='select * from source_table',
                    compiled=True,
                    extra_ctes_injected=False,
                    extra_ctes=[],
                    injected_sql='',
                    compiled_sql='select * from source_table'
                ),
            },
            docs={},
            generated_at='2018-02-14T09:15:13Z',
            disabled=[],
            files={},
        )

        result, output_graph = dbt.compilation.prepend_ctes(
            input_graph.nodes['model.root.view'],
            input_graph)

        self.assertEqual(result, input_graph.nodes['model.root.view'])
        self.assertTrue(result.extra_ctes_injected)
        self.assertEqualIgnoreWhitespace(
            result.injected_sql,
            ('with __dbt__CTE__ephemeral_level_two as ('
             'select * from source_table'
             '), __dbt__CTE__ephemeral as ('
             'select * from __dbt__CTE__ephemeral_level_two'
             ') '
             'select * from __dbt__CTE__ephemeral'))

        self.assertTrue(output_graph.nodes['model.root.ephemeral'].extra_ctes_injected)
        self.assertTrue(output_graph.nodes['model.root.ephemeral_level_two'].extra_ctes_injected)



from .utils import ContractTestCase

from hologram import ValidationError

from dbt.contracts.project import Project

class TestProject(ContractTestCase):
    ContractType = Project

    def test_minimal(self):
        dct = {
            'name': 'test',
            'version': '1.0',
            'profile': 'test',
            'project-root': '/usr/src/app',
        }
        project = Project(
            name='test',
            version='1.0',
            profile='test',
            project_root='/usr/src/app',
        )
        self.assert_from_dict(project, dct)

    def test_invalid_name(self):
        dct = {
            'name': 'log',
            'version': '1.0',
            'profile': 'test',
            'project-root': '/usr/src/app',
        }
        with self.assertRaises(ValidationError):
            Project.from_dict(dct)



from unittest import TestCase
from dbt.adapters.cache import RelationsCache
from dbt.adapters.base.relation import BaseRelation
from multiprocessing.dummy import Pool as ThreadPool
import dbt.exceptions

import random
import time


def make_relation(database, schema, identifier):
    return BaseRelation.create(database=database, schema=schema,
                               identifier=identifier)

def make_mock_relationship(database, schema, identifier):
    return BaseRelation.create(
        database=database, schema=schema, identifier=identifier, type='view'
    )


class TestCache(TestCase):
    def setUp(self):
        self.cache = RelationsCache()

    def assert_relations_state(self, database, schema, identifiers):
        relations = self.cache.get_relations(database, schema)
        for identifier, expect in identifiers.items():
            found = any(
                (r.identifier == identifier and \
                 r.schema == schema and \
                 r.database == database)
                for r in relations
            )
            msg = '{}.{}.{} was{} found in the cache!'.format(
                database, schema, identifier, '' if found else ' not'
            )
            self.assertEqual(expect, found, msg)

    def assert_relations_exist(self, database, schema, *identifiers):
        self.assert_relations_state(database, schema, {k: True for k in identifiers})

    def assert_relations_do_not_exist(self, database, schema, *identifiers):
        self.assert_relations_state(database, schema, {k: False for k in identifiers})


class TestEmpty(TestCache):
    def test_empty(self):
        self.assertEqual(len(self.cache.relations), 0)
        self.assertEqual(len(self.cache.get_relations('dbt', 'test')), 0)

class TestDrop(TestCache):
    def setUp(self):
        super().setUp()
        self.cache.add(make_relation('dbt', 'foo', 'bar'))

    def test_missing_identifier_ignored(self):
        self.cache.drop(make_relation('dbt', 'foo', 'bar1'))
        self.assert_relations_exist('dbt', 'foo', 'bar')
        self.assertEqual(len(self.cache.relations), 1)

    def test_missing_schema_ignored(self):
        self.cache.drop(make_relation('dbt', 'foo1', 'bar'))
        self.assert_relations_exist('dbt', 'foo', 'bar')
        self.assertEqual(len(self.cache.relations), 1)

    def test_missing_db_ignored(self):
        self.cache.drop(make_relation('dbt1', 'foo', 'bar'))
        self.assert_relations_exist('dbt', 'foo', 'bar')
        self.assertEqual(len(self.cache.relations), 1)

    def test_drop(self):
        self.cache.drop(make_relation('dbt', 'foo', 'bar'))
        self.assert_relations_do_not_exist('dbt', 'foo', 'bar')
        self.assertEqual(len(self.cache.relations), 0)


class TestAddLink(TestCache):
    def setUp(self):
        super().setUp()
        self.cache.add(make_relation('dbt', 'schema', 'foo'))
        self.cache.add(make_relation('dbt_2', 'schema', 'bar'))
        self.cache.add(make_relation('dbt', 'schema_2', 'bar'))

    def test_no_src(self):
        self.assert_relations_exist('dbt', 'schema', 'foo')
        self.assert_relations_do_not_exist('dbt', 'schema', 'bar')

        self.cache.add_link(make_relation('dbt', 'schema', 'bar'),
                            make_relation('dbt', 'schema', 'foo'))

        self.assert_relations_exist('dbt', 'schema', 'foo', 'bar')

    def test_no_dst(self):
        self.assert_relations_exist('dbt', 'schema', 'foo')
        self.assert_relations_do_not_exist('dbt', 'schema', 'bar')

        self.cache.add_link(make_relation('dbt', 'schema', 'foo'),
                            make_relation('dbt', 'schema', 'bar'))

        self.assert_relations_exist('dbt', 'schema', 'foo', 'bar')


class TestRename(TestCache):
    def setUp(self):
        super().setUp()
        self.cache.add(make_relation('DBT', 'schema', 'foo'))
        self.assert_relations_exist('DBT', 'schema', 'foo')
        self.assertEqual(self.cache.schemas, {('dbt', 'schema')})

    def test_no_source_error(self):
        # dest should be created anyway (it's probably a temp table)
        self.cache.rename(make_relation('DBT', 'schema', 'bar'),
                          make_relation('DBT', 'schema', 'baz'))

        self.assertEqual(len(self.cache.relations), 2)
        self.assert_relations_exist('DBT', 'schema', 'foo', 'baz')

    def test_dest_exists_error(self):
        foo = make_relation('DBT', 'schema', 'foo')
        bar = make_relation('DBT', 'schema', 'bar')
        self.cache.add(bar)
        self.assert_relations_exist('DBT', 'schema', 'foo', 'bar')

        with self.assertRaises(dbt.exceptions.InternalException):
            self.cache.rename(foo, bar)

        self.assert_relations_exist('DBT', 'schema', 'foo', 'bar')

    def test_dest_different_db(self):
        self.cache.rename(make_relation('DBT', 'schema', 'foo'),
                          make_relation('DBT_2', 'schema', 'foo'))
        self.assert_relations_exist('DBT_2', 'schema', 'foo')
        self.assert_relations_do_not_exist('DBT', 'schema', 'foo')
        self.assertEqual(self.cache.schemas, {('dbt_2', 'schema')})
        self.assertEqual(len(self.cache.relations), 1)

    def test_rename_identifier(self):
        self.cache.rename(make_relation('DBT', 'schema', 'foo'),
                          make_relation('DBT', 'schema', 'bar'))

        self.assert_relations_exist('DBT', 'schema', 'bar')
        self.assert_relations_do_not_exist('DBT', 'schema', 'foo')
        self.assertEqual(self.cache.schemas, {('dbt', 'schema')})

        relation = self.cache.relations[('dbt', 'schema', 'bar')]
        self.assertEqual(relation.inner.schema, 'schema')
        self.assertEqual(relation.inner.identifier, 'bar')
        self.assertEqual(relation.schema, 'schema')
        self.assertEqual(relation.identifier, 'bar')

    def test_rename_db(self):
        self.cache.rename(make_relation('DBT', 'schema', 'foo'),
                          make_relation('DBT_2', 'schema', 'foo'))

        self.assertEqual(len(self.cache.get_relations('DBT', 'schema')), 0)
        self.assertEqual(len(self.cache.get_relations('DBT_2', 'schema')), 1)
        self.assert_relations_exist('DBT_2', 'schema', 'foo')
        self.assert_relations_do_not_exist('DBT', 'schema', 'foo')
        self.assertEqual(self.cache.schemas, {('dbt_2', 'schema')})

        relation = self.cache.relations[('dbt_2', 'schema', 'foo')]
        self.assertEqual(relation.inner.database, 'DBT_2')
        self.assertEqual(relation.inner.schema, 'schema')
        self.assertEqual(relation.inner.identifier, 'foo')
        self.assertEqual(relation.database, 'dbt_2')
        self.assertEqual(relation.schema, 'schema')
        self.assertEqual(relation.identifier, 'foo')

    def test_rename_schema(self):
        self.cache.rename(make_relation('DBT', 'schema', 'foo'),
                          make_relation('DBT', 'schema_2', 'foo'))

        self.assertEqual(len(self.cache.get_relations('DBT', 'schema')), 0)
        self.assertEqual(len(self.cache.get_relations('DBT', 'schema_2')), 1)
        self.assert_relations_exist('DBT', 'schema_2', 'foo')
        self.assert_relations_do_not_exist('DBT', 'schema', 'foo')
        self.assertEqual(self.cache.schemas, {('dbt', 'schema_2')})

        relation = self.cache.relations[('dbt', 'schema_2', 'foo')]
        self.assertEqual(relation.inner.database, 'DBT')
        self.assertEqual(relation.inner.schema, 'schema_2')
        self.assertEqual(relation.inner.identifier, 'foo')
        self.assertEqual(relation.database, 'dbt')
        self.assertEqual(relation.schema, 'schema_2')
        self.assertEqual(relation.identifier, 'foo')


class TestGetRelations(TestCache):
    def setUp(self):
        super().setUp()
        self.relation = make_relation('dbt', 'foo', 'bar')
        self.cache.add(self.relation)

    def test_get_by_name(self):
        relations = self.cache.get_relations('dbt', 'foo')
        self.assertEqual(len(relations), 1)
        self.assertIs(relations[0], self.relation)

    def test_get_by_uppercase_schema(self):
        relations = self.cache.get_relations('dbt', 'FOO')
        self.assertEqual(len(relations), 1)
        self.assertIs(relations[0], self.relation)

    def test_get_by_uppercase_db(self):
        relations = self.cache.get_relations('DBT', 'foo')
        self.assertEqual(len(relations), 1)
        self.assertIs(relations[0], self.relation)

    def test_get_by_uppercase_schema_and_db(self):
        relations = self.cache.get_relations('DBT', 'FOO')
        self.assertEqual(len(relations), 1)
        self.assertIs(relations[0], self.relation)

    def test_get_by_wrong_db(self):
        relations = self.cache.get_relations('dbt_2', 'foo')
        self.assertEqual(len(relations), 0)

    def test_get_by_wrong_schema(self):
        relations = self.cache.get_relations('dbt', 'foo_2')
        self.assertEqual(len(relations), 0)


class TestAdd(TestCache):
    def setUp(self):
        super().setUp()
        self.relation = make_relation('dbt', 'foo', 'bar')
        self.cache.add(self.relation)

    def test_add(self):
        relations = self.cache.get_relations('dbt', 'foo')
        self.assertEqual(len(relations), 1)
        self.assertEqual(len(self.cache.relations), 1)
        self.assertIs(relations[0], self.relation)

    def test_add_twice(self):
        # add a new relation with same name
        self.cache.add(make_relation('dbt', 'foo', 'bar'))
        self.assertEqual(len(self.cache.relations), 1)
        self.assertEqual(self.cache.schemas, {('dbt', 'foo')})
        self.assert_relations_exist('dbt', 'foo', 'bar')

    def add_uppercase_schema(self):
        self.cache.add(make_relation('dbt', 'FOO', 'baz'))

        self.assertEqual(len(self.cache.relations), 2)
        relations = self.cache.get_relations('dbt', 'foo')
        self.assertEqual(len(relations), 2)
        self.assertEqual(self.cache.schemas, {('dbt', 'foo')})
        self.assertIsNot(self.cache.relations[('dbt', 'foo', 'bar')].inner, None)
        self.assertIsNot(self.cache.relations[('dbt', 'foo', 'baz')].inner, None)

    def add_different_db(self):
        self.cache.add(make_relation('dbt_2', 'foo', 'bar'))

        self.assertEqual(len(self.cache.relations), 2)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 1)
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 1)
        self.assertEqual(self.cache.schemas, {('dbt', 'foo'), ('dbt_2', 'foo')})
        self.assertIsNot(self.cache.relations[('dbt', 'foo', 'bar')].inner, None)
        self.assertIsNot(self.cache.relations[('dbt_2', 'foo', 'bar')].inner, None)


class TestLikeDbt(TestCase):
    def setUp(self):
        self.cache = RelationsCache()
        self._sleep = True

        # add a bunch of cache entries
        for ident in 'abcdef':
            self.cache.add(make_relation('dbt', 'schema', ident))
        # 'b' references 'a'
        self.cache.add_link(make_relation('dbt', 'schema', 'a'),
                            make_relation('dbt', 'schema', 'b'))
        # and 'c' references 'b'
        self.cache.add_link(make_relation('dbt', 'schema', 'b'),
                            make_relation('dbt', 'schema', 'c'))
        # and 'd' references 'b'
        self.cache.add_link(make_relation('dbt', 'schema', 'b'),
                            make_relation('dbt', 'schema', 'd'))
        # and 'e' references 'a'
        self.cache.add_link(make_relation('dbt', 'schema', 'a'),
                            make_relation('dbt', 'schema', 'e'))
        # and 'f' references 'd'
        self.cache.add_link(make_relation('dbt', 'schema', 'd'),
                            make_relation('dbt', 'schema', 'f'))
        # so drop propagation goes (a -> (b -> (c (d -> f))) e)

    def assert_has_relations(self, expected):
        current = set(r.identifier for r in self.cache.get_relations('dbt', 'schema'))
        self.assertEqual(current, expected)

    def test_drop_inner(self):
        self.assert_has_relations(set('abcdef'))
        self.cache.drop(make_relation('dbt', 'schema', 'b'))
        self.assert_has_relations({'a', 'e'})

    def test_rename_and_drop(self):
        self.assert_has_relations(set('abcdef'))
        # drop the backup/tmp
        self.cache.drop(make_relation('dbt', 'schema', 'b__backup'))
        self.cache.drop(make_relation('dbt', 'schema', 'b__tmp'))
        self.assert_has_relations(set('abcdef'))
        # create a new b__tmp
        self.cache.add(make_relation('dbt', 'schema', 'b__tmp',))
        self.assert_has_relations(set('abcdef') | {'b__tmp'})
        # rename b -> b__backup
        self.cache.rename(make_relation('dbt', 'schema', 'b'),
                          make_relation('dbt', 'schema', 'b__backup'))
        self.assert_has_relations(set('acdef') | {'b__tmp', 'b__backup'})
        # rename temp to b
        self.cache.rename(make_relation('dbt', 'schema', 'b__tmp'),
                          make_relation('dbt', 'schema', 'b'))
        self.assert_has_relations(set('abcdef') | {'b__backup'})

        # drop backup, everything that used to depend on b should be gone, but
        # b itself should still exist
        self.cache.drop(make_relation('dbt', 'schema', 'b__backup'))
        self.assert_has_relations(set('abe'))
        relation = self.cache.relations[('dbt', 'schema', 'a')]
        self.assertEqual(len(relation.referenced_by), 1)

    def _rand_sleep(self):
        if not self._sleep:
            return
        time.sleep(random.random() * 0.1)

    def _target(self, ident):
        self._rand_sleep()
        self.cache.rename(make_relation('dbt', 'schema', ident),
                          make_relation('dbt', 'schema', ident+'__backup'))
        self._rand_sleep()
        self.cache.add(make_relation('dbt', 'schema', ident+'__tmp')
        )
        self._rand_sleep()
        self.cache.rename(make_relation('dbt', 'schema', ident+'__tmp'),
                          make_relation('dbt', 'schema', ident))
        self._rand_sleep()
        self.cache.drop(make_relation('dbt', 'schema', ident+'__backup'))
        return ident, self.cache.get_relations('dbt', 'schema')

    def test_threaded(self):
        # add three more short subchains for threads to test on
        for ident in 'ghijklmno':
            obj = make_mock_relationship('test_db', 'schema', ident)
            self.cache.add(make_relation('dbt', 'schema', ident))

        self.cache.add_link(make_relation('dbt', 'schema', 'a'),
                            make_relation('dbt', 'schema', 'g'))
        self.cache.add_link(make_relation('dbt', 'schema', 'g'),
                            make_relation('dbt', 'schema', 'h'))
        self.cache.add_link(make_relation('dbt', 'schema', 'h'),
                            make_relation('dbt', 'schema', 'i'))

        self.cache.add_link(make_relation('dbt', 'schema', 'a'),
                            make_relation('dbt', 'schema', 'j'))
        self.cache.add_link(make_relation('dbt', 'schema', 'j'),
                            make_relation('dbt', 'schema', 'k'))
        self.cache.add_link(make_relation('dbt', 'schema', 'k'),
                            make_relation('dbt', 'schema', 'l'))

        self.cache.add_link(make_relation('dbt', 'schema', 'a'),
                            make_relation('dbt', 'schema', 'm'))
        self.cache.add_link(make_relation('dbt', 'schema', 'm'),
                            make_relation('dbt', 'schema', 'n'))
        self.cache.add_link(make_relation('dbt', 'schema', 'n'),
                            make_relation('dbt', 'schema', 'o'))

        pool = ThreadPool(4)
        results = list(pool.imap_unordered(self._target, ('b', 'g', 'j', 'm')))
        pool.close()
        pool.join()
        # at a minimum, we expect each table to "see" itself, its parent ('a'),
        # and the unrelated table ('a')
        min_expect = {
            'b': {'a', 'b', 'e'},
            'g': {'a', 'g', 'e'},
            'j': {'a', 'j', 'e'},
            'm': {'a', 'm', 'e'},
        }

        for ident, relations in results:
            seen = set(r.identifier for r in relations)
            self.assertTrue(min_expect[ident].issubset(seen))

        self.assert_has_relations(set('abgjme'))

    def test_threaded_repeated(self):
        for _ in range(10):
            self.setUp()
            self._sleep = False
            self.test_threaded()


class TestComplexCache(TestCase):
    def setUp(self):
        self.cache = RelationsCache()
        inputs = [
            ('dbt', 'foo', 'table1'),
            ('dbt', 'foo', 'table3'),
            ('dbt', 'foo', 'table4'),
            ('dbt', 'bar', 'table2'),
            ('dbt', 'bar', 'table3'),
            ('dbt_2', 'foo', 'table1'),
            ('dbt_2', 'foo', 'table2'),
        ]
        self.inputs = [make_relation(d, s, i) for d, s, i in inputs]
        for relation in self.inputs:
            self.cache.add(relation)

        # dbt.foo.table3 references dbt.foo.table1
        # (create view dbt.foo.table3 as (select * from dbt.foo.table1...))
        self.cache.add_link(
            make_relation('dbt', 'foo', 'table1'),
            make_relation('dbt', 'foo', 'table3')
        )
        # dbt.bar.table3 references dbt.foo.table3
        # (create view dbt.bar.table5 as (select * from dbt.foo.table3...))
        self.cache.add_link(
            make_relation('dbt', 'foo', 'table3'),
            make_relation('dbt', 'bar', 'table3')
        )

        # dbt.foo.table4 also references dbt.foo.table1
        self.cache.add_link(
            make_relation('dbt', 'foo', 'table1'),
            make_relation('dbt', 'foo', 'table4')
        )

        # and dbt_2.foo.table1 references dbt.foo.table1
        self.cache.add_link(
            make_relation('dbt', 'foo', 'table1'),
            make_relation('dbt_2', 'foo', 'table1'),
        )

    def test_get_relations(self):
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 3)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 2)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 2)
        self.assertEqual(len(self.cache.relations), 7)

    def test_drop_one(self):
        # dropping dbt.bar.table2 should only drop itself
        self.cache.drop(make_relation('dbt', 'bar', 'table2'))
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 3)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 1)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 2)
        self.assertEqual(len(self.cache.relations), 6)

    def test_drop_many(self):
        # dropping dbt.foo.table1 should drop everything but dbt.bar.table2 and
        # dbt_2.foo.table2
        self.cache.drop(make_relation('dbt', 'foo', 'table1'))
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 0)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 1)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 1)
        self.assertEqual(len(self.cache.relations), 2)

    def test_rename_root(self):
        self.cache.rename(make_relation('dbt', 'foo', 'table1'),
                          make_relation('dbt', 'bar', 'table1'))
        retrieved = self.cache.relations[('dbt', 'bar', 'table1')].inner
        self.assertEqual(retrieved.schema, 'bar')
        self.assertEqual(retrieved.identifier, 'table1')
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 2)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 3)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 2)
        self.assertEqual(len(self.cache.relations), 7)

        # make sure drops still cascade from the renamed table
        self.cache.drop(make_relation('dbt', 'bar', 'table1'))
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 0)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 1)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 1)
        self.assertEqual(len(self.cache.relations), 2)

    def test_rename_branch(self):
        self.cache.rename(make_relation('dbt', 'foo', 'table3'),
                          make_relation('dbt', 'foo', 'table2'))
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 3)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 2)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 2)

        # make sure drops still cascade through the renamed table
        self.cache.drop(make_relation('dbt', 'foo', 'table1'))
        self.assertEqual(len(self.cache.get_relations('dbt', 'foo')), 0)
        self.assertEqual(len(self.cache.get_relations('dbt', 'bar')), 1)
        self.assertEqual(len(self.cache.get_relations('dbt_2', 'foo')), 1)
        self.assertEqual(len(self.cache.relations), 2)



import os
import tempfile
import unittest
from unittest import mock

from dbt import linker
try:
    from queue import Empty
except ImportError:
    from Queue import Empty


def _mock_manifest(nodes):
    manifest = mock.MagicMock(nodes={
        n: mock.MagicMock(unique_id=n) for n in nodes
    })
    manifest.expect.side_effect = lambda n: mock.MagicMock(unique_id=n)
    return manifest


class LinkerTest(unittest.TestCase):

    def setUp(self):
        self.patcher = mock.patch.object(linker, 'is_blocking_dependency')
        self.is_blocking_dependency = self.patcher.start()
        self.is_blocking_dependency.return_value = True
        self.linker = linker.Linker()

    def tearDown(self):
        self.patcher.stop()

    def test_linker_add_node(self):
        expected_nodes = ['A', 'B', 'C']
        for node in expected_nodes:
            self.linker.add_node(node)

        actual_nodes = self.linker.nodes()
        for node in expected_nodes:
            self.assertIn(node, actual_nodes)

        self.assertEqual(len(actual_nodes), len(expected_nodes))

    def test_linker_write_and_read_graph(self):
        expected_nodes = ['A', 'B', 'C']
        for node in expected_nodes:
            self.linker.add_node(node)

        manifest = _mock_manifest('ABC')
        (fd, fname) = tempfile.mkstemp()
        os.close(fd)
        try:
            self.linker.write_graph(fname, manifest)
            new_linker = linker.from_file(fname)
        finally:
            os.unlink(fname)

        actual_nodes = new_linker.nodes()
        for node in expected_nodes:
            self.assertIn(node, actual_nodes)

        self.assertEqual(len(actual_nodes), len(expected_nodes))

    def assert_would_join(self, queue):
        """test join() without timeout risk"""
        self.assertEqual(queue.inner.unfinished_tasks, 0)

    def test_linker_add_dependency(self):
        actual_deps = [('A', 'B'), ('A', 'C'), ('B', 'C')]

        for (l, r) in actual_deps:
            self.linker.dependency(l, r)

        manifest = _mock_manifest('ABC')
        queue = self.linker.as_graph_queue(manifest)

        got = queue.get(block=False)
        self.assertEqual(got.unique_id, 'C')
        with self.assertRaises(Empty):
            queue.get(block=False)
        self.assertFalse(queue.empty())
        queue.mark_done('C')
        self.assertFalse(queue.empty())

        got = queue.get(block=False)
        self.assertEqual(got.unique_id, 'B')
        with self.assertRaises(Empty):
            queue.get(block=False)
        self.assertFalse(queue.empty())
        queue.mark_done('B')
        self.assertFalse(queue.empty())

        got = queue.get(block=False)
        self.assertEqual(got.unique_id, 'A')
        with self.assertRaises(Empty):
            queue.get(block=False)
        self.assertTrue(queue.empty())
        queue.mark_done('A')
        self.assert_would_join(queue)
        self.assertTrue(queue.empty())

    def test_linker_add_disjoint_dependencies(self):
        actual_deps = [('A', 'B')]
        additional_node = 'Z'

        for (l, r) in actual_deps:
            self.linker.dependency(l, r)
        self.linker.add_node(additional_node)


        manifest = _mock_manifest('ABZ')
        queue = self.linker.as_graph_queue(manifest)

        # the first one we get must be B, it has the longest dep chain
        first = queue.get(block=False)
        self.assertEqual(first.unique_id, 'B')
        self.assertFalse(queue.empty())
        queue.mark_done('B')
        self.assertFalse(queue.empty())

        second = queue.get(block=False)
        self.assertIn(second.unique_id, {'A', 'Z'})
        self.assertFalse(queue.empty())
        queue.mark_done(second.unique_id)
        self.assertFalse(queue.empty())

        third = queue.get(block=False)
        self.assertIn(third.unique_id, {'A', 'Z'})
        with self.assertRaises(Empty):
            queue.get(block=False)
        self.assertNotEqual(second.unique_id, third.unique_id)
        self.assertTrue(queue.empty())
        queue.mark_done(third.unique_id)
        self.assert_would_join(queue)
        self.assertTrue(queue.empty())

    def test_linker_dependencies_limited_to_some_nodes(self):
        actual_deps = [('A', 'B'), ('B', 'C'), ('C', 'D')]

        for (l, r) in actual_deps:
            self.linker.dependency(l, r)

        queue = self.linker.as_graph_queue(_mock_manifest('ABCD'), ['B'])
        got = queue.get(block=False)
        self.assertEqual(got.unique_id, 'B')
        self.assertTrue(queue.empty())
        queue.mark_done('B')
        self.assert_would_join(queue)

        queue_2 = self.linker.as_graph_queue(_mock_manifest('ABCD'), ['A', 'B'])
        got = queue_2.get(block=False)
        self.assertEqual(got.unique_id, 'B')
        self.assertFalse(queue_2.empty())
        with self.assertRaises(Empty):
            queue_2.get(block=False)
        queue_2.mark_done('B')
        self.assertFalse(queue_2.empty())

        got = queue_2.get(block=False)
        self.assertEqual(got.unique_id, 'A')
        self.assertTrue(queue_2.empty())
        with self.assertRaises(Empty):
            queue_2.get(block=False)
        self.assertTrue(queue_2.empty())
        queue_2.mark_done('A')
        self.assert_would_join(queue_2)

    def test_linker_bad_limit_throws_runtime_error(self):
        actual_deps = [('A', 'B'), ('B', 'C'), ('C', 'D')]

        for (l, r) in actual_deps:
            self.linker.dependency(l, r)

        with self.assertRaises(RuntimeError):
            self.linker.as_graph_queue(_mock_manifest('ABCD'), ['ZZZ'])

    def test__find_cycles__cycles(self):
        actual_deps = [('A', 'B'), ('B', 'C'), ('C', 'A')]

        for (l, r) in actual_deps:
            self.linker.dependency(l, r)

        self.assertIsNotNone(self.linker.find_cycles())

    def test__find_cycles__no_cycles(self):
        actual_deps = [('A', 'B'), ('B', 'C'), ('C', 'D')]

        for (l, r) in actual_deps:
            self.linker.dependency(l, r)

        self.assertIsNone(self.linker.find_cycles())



import unittest

from dbt.clients.jinja import get_template
from dbt.clients.jinja import extract_toplevel_blocks
from dbt.exceptions import CompilationException


class TestJinja(unittest.TestCase):
    def test_do(self):
        s = '{% set my_dict = {} %}\n{% do my_dict.update(a=1) %}'

        template = get_template(s, {})
        mod = template.make_module()
        self.assertEqual(mod.my_dict, {'a': 1})


class TestBlockLexer(unittest.TestCase):
    def test_basic(self):
        body = '{{ config(foo="bar") }}\r\nselect * from this.that\r\n'
        block_data = '  \n\r\t{%- mytype foo %}'+body+'{%endmytype -%}'
        blocks = extract_toplevel_blocks(block_data, allowed_blocks={'mytype'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'mytype')
        self.assertEqual(blocks[0].block_name, 'foo')
        self.assertEqual(blocks[0].contents, body)
        self.assertEqual(blocks[0].full_block, block_data)

    def test_multiple(self):
        body_one = '{{ config(foo="bar") }}\r\nselect * from this.that\r\n'
        body_two = (
            '{{ config(bar=1)}}\r\nselect * from {% if foo %} thing '
            '{% else %} other_thing {% endif %}'
        )

        block_data = (
            '  {% mytype foo %}' + body_one + '{% endmytype %}' +
            '\r\n{% othertype bar %}' + body_two + '{% endothertype %}'
        )
        blocks = extract_toplevel_blocks(block_data, allowed_blocks={'mytype', 'othertype'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 2)

    def test_comments(self):
        body = '{{ config(foo="bar") }}\r\nselect * from this.that\r\n'
        comment = '{# my comment #}'
        block_data = '  \n\r\t{%- mytype foo %}'+body+'{%endmytype -%}'
        blocks = extract_toplevel_blocks(comment+block_data, allowed_blocks={'mytype'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'mytype')
        self.assertEqual(blocks[0].block_name, 'foo')
        self.assertEqual(blocks[0].contents, body)
        self.assertEqual(blocks[0].full_block, block_data)

    def test_evil_comments(self):
        body = '{{ config(foo="bar") }}\r\nselect * from this.that\r\n'
        comment = '{# external comment {% othertype bar %} select * from thing.other_thing{% endothertype %} #}'
        block_data = '  \n\r\t{%- mytype foo %}'+body+'{%endmytype -%}'
        blocks = extract_toplevel_blocks(comment+block_data, allowed_blocks={'mytype'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'mytype')
        self.assertEqual(blocks[0].block_name, 'foo')
        self.assertEqual(blocks[0].contents, body)
        self.assertEqual(blocks[0].full_block, block_data)

    def test_nested_comments(self):
        body = '{# my comment #} {{ config(foo="bar") }}\r\nselect * from {# my other comment embedding {% endmytype %} #} this.that\r\n'
        block_data = '  \n\r\t{%- mytype foo %}'+body+'{% endmytype -%}'
        comment = '{# external comment {% othertype bar %} select * from thing.other_thing{% endothertype %} #}'
        blocks = extract_toplevel_blocks(comment+block_data, allowed_blocks={'mytype'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'mytype')
        self.assertEqual(blocks[0].block_name, 'foo')
        self.assertEqual(blocks[0].contents, body)
        self.assertEqual(blocks[0].full_block, block_data)

    def test_complex_file(self):
        blocks = extract_toplevel_blocks(complex_snapshot_file, allowed_blocks={'mytype', 'myothertype'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 3)
        self.assertEqual(blocks[0].block_type_name, 'mytype')
        self.assertEqual(blocks[0].block_name, 'foo')
        self.assertEqual(blocks[0].full_block, '{% mytype foo %} some stuff {% endmytype %}')
        self.assertEqual(blocks[0].contents, ' some stuff ')
        self.assertEqual(blocks[1].block_type_name, 'mytype')
        self.assertEqual(blocks[1].block_name, 'bar')
        self.assertEqual(blocks[1].full_block, bar_block)
        self.assertEqual(blocks[1].contents, bar_block[16:-15].rstrip())
        self.assertEqual(blocks[2].block_type_name, 'myothertype')
        self.assertEqual(blocks[2].block_name, 'x')
        self.assertEqual(blocks[2].full_block, x_block.strip())
        self.assertEqual(blocks[2].contents, x_block[len('\n{% myothertype x %}'):-len('{% endmyothertype %}\n')])

    def test_peaceful_macro_coexistence(self):
        body = '{# my macro #} {% macro foo(a, b) %} do a thing {%- endmacro %} {# my model #} {% a b %} test {% enda %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'macro', 'a'}, collect_raw_data=True)
        self.assertEqual(len(blocks), 4)
        self.assertEqual(blocks[0].full_block, '{# my macro #} ')
        self.assertEqual(blocks[1].block_type_name, 'macro')
        self.assertEqual(blocks[1].block_name, 'foo')
        self.assertEqual(blocks[1].contents, ' do a thing')
        self.assertEqual(blocks[2].full_block, ' {# my model #} ')
        self.assertEqual(blocks[3].block_type_name, 'a')
        self.assertEqual(blocks[3].block_name, 'b')
        self.assertEqual(blocks[3].contents, ' test ')

    def test_macro_with_trailing_data(self):
        body = '{# my macro #} {% macro foo(a, b) %} do a thing {%- endmacro %} {# my model #} {% a b %} test {% enda %} raw data so cool'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'macro', 'a'}, collect_raw_data=True)
        self.assertEqual(len(blocks), 5)
        self.assertEqual(blocks[0].full_block, '{# my macro #} ')
        self.assertEqual(blocks[1].block_type_name, 'macro')
        self.assertEqual(blocks[1].block_name, 'foo')
        self.assertEqual(blocks[1].contents, ' do a thing')
        self.assertEqual(blocks[2].full_block, ' {# my model #} ')
        self.assertEqual(blocks[3].block_type_name, 'a')
        self.assertEqual(blocks[3].block_name, 'b')
        self.assertEqual(blocks[3].contents, ' test ')
        self.assertEqual(blocks[4].full_block, ' raw data so cool')

    def test_macro_with_crazy_args(self):
        body = '''{% macro foo(a, b=asdf("cool this is 'embedded'" * 3) + external_var, c)%}cool{# block comment with {% endmacro %} in it #} stuff here {% endmacro %}'''
        blocks = extract_toplevel_blocks(body, allowed_blocks={'macro'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'macro')
        self.assertEqual(blocks[0].block_name, 'foo')
        self.assertEqual(blocks[0].contents, 'cool{# block comment with {% endmacro %} in it #} stuff here ')

    def test_materialization_parse(self):
        body = '{% materialization xxx, default %} ... {% endmaterialization %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'materialization'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'materialization')
        self.assertEqual(blocks[0].block_name, 'xxx')
        self.assertEqual(blocks[0].full_block, body)

        body = '{% materialization xxx, adapter="other" %} ... {% endmaterialization %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'materialization'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'materialization')
        self.assertEqual(blocks[0].block_name, 'xxx')
        self.assertEqual(blocks[0].full_block, body)

    def test_nested_not_ok(self):
        # we don't allow nesting same blocks
        body = '{% myblock a %} {% myblock b %} {% endmyblock %} {% endmyblock %}'
        with self.assertRaises(CompilationException):
            extract_toplevel_blocks(body, allowed_blocks={'myblock'})

    def test_incomplete_block_failure(self):
        fullbody = '{% myblock foo %} {% endmyblock %}'
        for length in range(len('{% myblock foo %}'), len(fullbody)-1):
            body = fullbody[:length]
            with self.assertRaises(CompilationException):
                extract_toplevel_blocks(body, allowed_blocks={'myblock'})

    def test_wrong_end_failure(self):
        body = '{% myblock foo %} {% endotherblock %}'
        with self.assertRaises(CompilationException):
            extract_toplevel_blocks(body, allowed_blocks={'myblock', 'otherblock'})

    def test_comment_no_end_failure(self):
        body = '{# '
        with self.assertRaises(CompilationException):
            extract_toplevel_blocks(body)

    def test_comment_only(self):
        body = '{# myblock #}'
        blocks = extract_toplevel_blocks(body)
        self.assertEqual(len(blocks), 1)
        blocks = extract_toplevel_blocks(body, collect_raw_data=False)
        self.assertEqual(len(blocks), 0)

    def test_comment_block_self_closing(self):
        # test the case where a comment start looks a lot like it closes itself
        # (but it doesn't in jinja!)
        body = '{#} {% myblock foo %} {#}'
        blocks = extract_toplevel_blocks(body, collect_raw_data=False)
        self.assertEqual(len(blocks), 0)

    def test_embedded_self_closing_comment_block(self):
        body = '{% myblock foo %} {#}{% endmyblock %} {#}{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, body)
        self.assertEqual(blocks[0].contents, ' {#}{% endmyblock %} {#}')

    def test_set_statement(self):
        body = '{% set x = 1 %}{% myblock foo %}hi{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, '{% myblock foo %}hi{% endmyblock %}')

    def test_set_block(self):
        body = '{% set x %}1{% endset %}{% myblock foo %}hi{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, '{% myblock foo %}hi{% endmyblock %}')

    def test_crazy_set_statement(self):
        body = '{% set x = (thing("{% myblock foo %}")) %}{% otherblock bar %}x{% endotherblock %}{% set y = otherthing("{% myblock foo %}") %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'otherblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, '{% otherblock bar %}x{% endotherblock %}')
        self.assertEqual(blocks[0].block_type_name, 'otherblock')

    def test_do_statement(self):
        body = '{% do thing.update() %}{% myblock foo %}hi{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, '{% myblock foo %}hi{% endmyblock %}')

    def test_deceptive_do_statement(self):
        body = '{% do thing %}{% myblock foo %}hi{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, '{% myblock foo %}hi{% endmyblock %}')

    def test_do_block(self):
        body = '{% do %}thing.update(){% enddo %}{% myblock foo %}hi{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'do', 'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 2)
        self.assertEqual(blocks[0].contents, 'thing.update()')
        self.assertEqual(blocks[0].block_type_name, 'do')
        self.assertEqual(blocks[1].full_block, '{% myblock foo %}hi{% endmyblock %}')

    def test_crazy_do_statement(self):
        body = '{% do (thing("{% myblock foo %}")) %}{% otherblock bar %}x{% endotherblock %}{% do otherthing("{% myblock foo %}") %}{% myblock x %}hi{% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock', 'otherblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 2)
        self.assertEqual(blocks[0].full_block, '{% otherblock bar %}x{% endotherblock %}')
        self.assertEqual(blocks[0].block_type_name, 'otherblock')
        self.assertEqual(blocks[1].full_block, '{% myblock x %}hi{% endmyblock %}')
        self.assertEqual(blocks[1].block_type_name, 'myblock')

    def test_awful_jinja(self):
        blocks = extract_toplevel_blocks(
                if_you_do_this_you_are_awful,
                allowed_blocks={'snapshot', 'materialization'},
                collect_raw_data=False
        )
        self.assertEqual(len(blocks), 2)
        self.assertEqual(len([b for b in blocks if b.block_type_name == '__dbt__data']), 0)
        self.assertEqual(blocks[0].block_type_name, 'snapshot')
        self.assertEqual(blocks[0].contents, '\n    '.join([
            '''{% set x = ("{% endsnapshot %}" + (40 * '%})')) %}''',
            '{# {% endsnapshot %} #}',
            '{% embedded %}',
            '    some block data right here',
            '{% endembedded %}'
        ]))
        self.assertEqual(blocks[1].block_type_name, 'materialization')
        self.assertEqual(blocks[1].contents, '\nhi\n')

    def test_quoted_endblock_within_block(self):
        body = '{% myblock something -%}  {% set x = ("{% endmyblock %}") %}  {% endmyblock %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'myblock'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'myblock')
        self.assertEqual(blocks[0].contents, '{% set x = ("{% endmyblock %}") %}  ')

    def test_docs_block(self):
        body = '{% docs __my_doc__ %} asdf {# nope {% enddocs %}} #} {% enddocs %} {% docs __my_other_doc__ %} asdf "{% enddocs %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'docs'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 2)
        self.assertEqual(blocks[0].block_type_name, 'docs')
        self.assertEqual(blocks[0].contents, ' asdf {# nope {% enddocs %}} #} ')
        self.assertEqual(blocks[0].block_name, '__my_doc__')
        self.assertEqual(blocks[1].block_type_name, 'docs')
        self.assertEqual(blocks[1].contents, ' asdf "')
        self.assertEqual(blocks[1].block_name, '__my_other_doc__')

    def test_docs_block_expr(self):
        body = '{% docs more_doc %} asdf {{ "{% enddocs %}" ~ "}}" }}{% enddocs %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'docs'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'docs')
        self.assertEqual(blocks[0].contents, ' asdf {{ "{% enddocs %}" ~ "}}" }}')
        self.assertEqual(blocks[0].block_name, 'more_doc')

    def test_unclosed_model_quotes(self):
        # test case for https://github.com/fishtown-analytics/dbt/issues/1533
        body = '{% model my_model -%} select * from "something"."something_else{% endmodel %}'
        blocks = extract_toplevel_blocks(body, allowed_blocks={'model'}, collect_raw_data=False)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].block_type_name, 'model')
        self.assertEqual(blocks[0].contents, 'select * from "something"."something_else')
        self.assertEqual(blocks[0].block_name, 'my_model')

    def test_if(self):
        # if you conditionally define your macros/models, don't
        body = '{% if true %}{% macro my_macro() %} adsf {% endmacro %}{% endif %}'
        with self.assertRaises(CompilationException):
            extract_toplevel_blocks(body)

    def test_if_innocuous(self):
        body = '{% if true %}{% something %}asdfasd{% endsomething %}{% endif %}'
        blocks = extract_toplevel_blocks(body)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, body)

    def test_for(self):
        # no for-loops over macros.
        body = '{% for x in range(10) %}{% macro my_macro() %} adsf {% endmacro %}{% endfor %}'
        with self.assertRaises(CompilationException):
            extract_toplevel_blocks(body)

    def test_for_innocuous(self):
        # no for-loops over macros.
        body = '{% for x in range(10) %}{% something my_something %} adsf {% endsomething %}{% endfor %}'
        blocks = extract_toplevel_blocks(body)
        self.assertEqual(len(blocks), 1)
        self.assertEqual(blocks[0].full_block, body)

    def test_endif(self):
        body = '{% snapshot foo %}select * from thing{% endsnapshot%}{% endif %}'
        with self.assertRaises(CompilationException) as err:
            extract_toplevel_blocks(body)
        self.assertIn('Got an unexpected control flow end tag, got endif but never saw a preceeding if (@ 53)', str(err.exception))

    def test_if_endfor(self):
        body = '{% if x %}...{% endfor %}{% endif %}'
        with self.assertRaises(CompilationException) as err:
            extract_toplevel_blocks(body)
        self.assertIn('Got an unexpected control flow end tag, got endfor but expected endif next (@ 13)', str(err.exception))


bar_block = '''{% mytype bar %}
{# a comment
    that inside it has
    {% mytype baz %}
{% endmyothertype %}
{% endmytype %}
{% endmytype %}
    {#
{% endmytype %}#}

some other stuff

{%- endmytype%}'''

x_block = '''
{% myothertype x %}
before
{##}
and after
{% endmyothertype %}
'''

complex_snapshot_file = '''
{#some stuff {% mytype foo %} #}
{% mytype foo %} some stuff {% endmytype %}

'''+bar_block+x_block


if_you_do_this_you_are_awful = '''
{#} here is a comment with a block inside {% block x %} asdf {% endblock %} {#}
{% do
    set('foo="bar"')
%}
{% set x = ("100" + "hello'" + '%}') %}
{% snapshot something -%}
    {% set x = ("{% endsnapshot %}" + (40 * '%})')) %}
    {# {% endsnapshot %} #}
    {% embedded %}
        some block data right here
    {% endembedded %}
{%- endsnapshot %}

{% raw %}
    {% set x = SYNTAX ERROR}
{% endraw %}


{% materialization whatever, adapter='thing' %}
hi
{% endmaterialization %}
'''



import pickle

from dbt.node_types import NodeType
from dbt.contracts.graph.parsed import (
    ParsedModelNode, DependsOn, NodeConfig, ColumnInfo, Hook, ParsedTestNode,
    TestConfig, ParsedSnapshotNode, TimestampSnapshotConfig, All, Docref,
    GenericSnapshotConfig, CheckSnapshotConfig, SnapshotStrategy,
    IntermediateSnapshotNode, ParsedNodePatch, ParsedMacro,
    MacroDependsOn, ParsedSourceDefinition, ParsedDocumentation, ParsedHookNode
)
from dbt.contracts.graph.unparsed import Quoting, FreshnessThreshold

from hologram import ValidationError
from .utils import ContractTestCase


class TestNodeConfig(ContractTestCase):
    ContractType = NodeConfig

    def test_basics(self):
        cfg_dict = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
        }
        cfg = self.ContractType()
        self.assert_symmetric(cfg, cfg_dict)

    def test_populated(self):
        cfg_dict = {
            'column_types': {'a': 'text'},
            'enabled': True,
            'materialized': 'table',
            'persist_docs': {},
            'post-hook': [{'sql': 'insert into blah(a, b) select "1", 1', 'transaction': True}],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'extra': 'even more',
        }
        cfg = self.ContractType(
            column_types={'a': 'text'},
            materialized='table',
            post_hook=[Hook(sql='insert into blah(a, b) select "1", 1')]
        )
        cfg._extra['extra'] = 'even more'

        self.assert_symmetric(cfg, cfg_dict)
        pickle.loads(pickle.dumps(cfg))


class TestParsedModelNode(ContractTestCase):
    ContractType = ParsedModelNode

    def test_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertTrue(node.is_refable)
        self.assertFalse(node.is_ephemeral)
        self.assertEqual(node.local_vars(), {})

        minimum = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_from_dict(node, minimum)
        pickle.loads(pickle.dumps(node))

    def test_complex(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("bar") }}',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': ['model.test.bar']},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': ['tag'],
            'config': {
                'column_types': {'a': 'text'},
                'enabled': True,
                'materialized': 'ephemeral',
                'persist_docs': {},
                'post-hook': [{'sql': 'insert into blah(a, b) select "1", 1', 'transaction': True}],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {'foo': 100},
            },
            'docrefs': [],
            'columns': {'a': {'name': 'a', 'description': 'a text field'}},
        }

        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from {{ ref("bar") }}',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(nodes=['model.test.bar']),
            description='My parsed node',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=['tag'],
            config=NodeConfig(
                column_types={'a': 'text'},
                materialized='ephemeral',
                post_hook=[Hook(sql='insert into blah(a, b) select "1", 1')],
                vars={'foo': 100},
            ),
            columns={'a': ColumnInfo('a', 'a text field')},
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertTrue(node.is_refable)
        self.assertTrue(node.is_ephemeral)
        self.assertEqual(node.local_vars(), {'foo': 100})

    def test_invalid_bad_tags(self):
        # bad top-level field
        bad_tags = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': 100,
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': None,
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
        }
        self.assert_fails_validation(bad_tags)

    def test_invalid_bad_materialized(self):
        # bad nested field
        bad_materialized = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': ['tag'],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': None,
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
        }
        self.assert_fails_validation(bad_materialized)

    def test_patch_ok(self):
        initial = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
        )
        patch = ParsedNodePatch(
            name='foo',
            description='The foo model',
            original_file_path='/path/to/schema.yml',
            columns={'a': ColumnInfo(name='a', description='a text field')},
            docrefs=[
                Docref(documentation_name='foo', documentation_package='test'),
            ],
        )

        initial.patch(patch)

        expected_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'The foo model',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'patch_path': '/path/to/schema.yml',
            'columns': {'a': {'name': 'a', 'description': 'a text field'}},
            'docrefs': [
                {
                    'documentation_name': 'foo',
                    'documentation_package': 'test',
                }
            ],
        }

        expected = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='The foo model',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
            patch_path='/path/to/schema.yml',
            columns={'a': ColumnInfo(name='a', description='a text field')},
            docrefs=[
                Docref(documentation_name='foo', documentation_package='test'),
            ],
        )
        self.assert_symmetric(expected, expected_dict)  # sanity check
        self.assertEqual(initial, expected)
        self.assert_symmetric(initial, expected_dict)

    def patch_invalid(self):
        initial = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
        )
        # invalid patch: description can't be None
        patch = ParsedNodePatch(
            name='foo',
            description=None,
            original_file_path='/path/to/schema.yml',
            columns={},
            docrefs=[],
        )
        with self.assertRaises(ValidationError):
            initial.patch(patch)


class TestParsedHookNode(ContractTestCase):
    ContractType = ParsedHookNode

    def test_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Operation),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Operation,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertFalse(node.is_refable)
        self.assertEqual(node.get_materialization(), 'view')

        minimum = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Operation),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_from_dict(node, minimum)
        pickle.loads(pickle.dumps(node))

    def test_complex(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Operation),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("bar") }}',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': ['model.test.bar']},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': ['tag'],
            'config': {
                'column_types': {'a': 'text'},
                'enabled': True,
                'materialized': 'table',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {'a': {'name': 'a', 'description': 'a text field'}},
            'index': 13,
        }

        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from {{ ref("bar") }}',
            name='foo',
            resource_type=NodeType.Operation,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(nodes=['model.test.bar']),
            description='My parsed node',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=['tag'],
            config=NodeConfig(
                column_types={'a': 'text'},
                materialized='table',
                post_hook=[]
            ),
            columns={'a': ColumnInfo('a', 'a text field')},
            index=13,
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertFalse(node.is_refable)
        self.assertEqual(node.get_materialization(), 'table')

    def test_invalid_index_type(self):
        # bad top-level field
        bad_index = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Operation),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': None,
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
            'index': 'a string!?',
        }
        self.assert_fails_validation(bad_index)


class TestParsedTestNode(ContractTestCase):
    ContractType = ParsedTestNode

    def test_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'test.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severity': 'error',
            },
            'docrefs': [],
            'columns': {},
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Test,
            unique_id='test.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=TestConfig(),
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertFalse(node.is_ephemeral)
        self.assertFalse(node.is_refable)
        self.assertEqual(node.get_materialization(), 'view')

        minimum = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'test.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_from_dict(node, minimum)
        pickle.loads(pickle.dumps(node))

    def test_complex(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("bar") }}',
            'unique_id': 'test.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': ['model.test.bar']},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': ['tag'],
            'config': {
                'column_types': {'a': 'text'},
                'enabled': True,
                'materialized': 'table',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severity': 'WARN',
                'extra_key': 'extra value'
            },
            'docrefs': [],
            'columns': {'a': {'name': 'a', 'description': 'a text field'}},
            'column_name': 'id',
        }

        cfg = TestConfig(
            column_types={'a': 'text'},
            materialized='table',
            severity='WARN'
        )
        cfg._extra.update({'extra_key': 'extra value'})

        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from {{ ref("bar") }}',
            name='foo',
            resource_type=NodeType.Test,
            unique_id='test.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(nodes=['model.test.bar']),
            description='My parsed node',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=['tag'],
            config=cfg,
            columns={'a': ColumnInfo('a', 'a text field')},
            column_name='id',
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)

    def test_invalid_column_name_type(self):
        # bad top-level field
        bad_column_name = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': 100,
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': None,
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severity': 'ERROR',
            },
            'docrefs': [],
            'columns': {},
            'column_name': {},
        }
        self.assert_fails_validation(bad_column_name)

    def test_invalid_missing_severity(self):
        # note the typo ('severtiy')
        missing_config_value = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': ['tag'],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': None,
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severtiy': 'WARN',
            },
            'docrefs': [],
            'columns': {},
        }
        self.assert_fails_validation(missing_config_value)

    def test_invalid_severity(self):
        invalid_config_value = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': 'My parsed node',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': ['tag'],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': None,
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severity': 'WERROR',  # invalid severity
            },
            'docrefs': [],
            'columns': {},
        }
        self.assert_fails_validation(invalid_config_value)


class TestTimestampSnapshotConfig(ContractTestCase):
    ContractType = TimestampSnapshotConfig

    def test_basics(self):
        cfg_dict = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'unique_key': 'id',
            'strategy': 'timestamp',
            'updated_at': 'last_update',
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
        }
        cfg = self.ContractType(
            strategy=SnapshotStrategy.Timestamp,
            updated_at='last_update',
            unique_key='id',
            target_database='some_snapshot_db',
            target_schema='some_snapshot_schema',
        )
        self.assert_symmetric(cfg, cfg_dict)
        pickle.loads(pickle.dumps(cfg))

    def test_populated(self):
        cfg_dict = {
            'column_types': {'a': 'text'},
            'enabled': True,
            'materialized': 'table',
            'persist_docs': {},
            'post-hook': [{'sql': 'insert into blah(a, b) select "1", 1', 'transaction': True}],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'extra': 'even more',
            'strategy': 'timestamp',
            'updated_at': 'last_update',
        }
        cfg = self.ContractType(
            column_types={'a': 'text'},
            materialized='table',
            post_hook=[Hook(sql='insert into blah(a, b) select "1", 1')],
            strategy=SnapshotStrategy.Timestamp,
            target_database='some_snapshot_db',
            target_schema='some_snapshot_schema',
            updated_at='last_update',
            unique_key='id',
        )
        cfg._extra['extra'] = 'even more'

        self.assert_symmetric(cfg, cfg_dict)

    def test_invalid_wrong_strategy(self):
        bad_type = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'strategy': 'check',
            'updated_at': 'last_update',
        }
        self.assert_fails_validation(bad_type)

    def test_invalid_missing_updated_at(self):
        bad_fields = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'strategy': 'timestamp',
            'check_cols': 'all'
        }
        self.assert_fails_validation(bad_fields)


class TestCheckSnapshotConfig(ContractTestCase):
    ContractType = CheckSnapshotConfig

    def test_basics(self):
        cfg_dict = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'strategy': 'check',
            'check_cols': 'all',
        }
        cfg = self.ContractType(
            strategy=SnapshotStrategy.Check,
            check_cols=All.All,
            unique_key='id',
            target_database='some_snapshot_db',
            target_schema='some_snapshot_schema',
        )
        self.assert_symmetric(cfg, cfg_dict)
        pickle.loads(pickle.dumps(cfg))

    def test_populated(self):
        cfg_dict = {
            'column_types': {'a': 'text'},
            'enabled': True,
            'materialized': 'table',
            'persist_docs': {},
            'post-hook': [{'sql': 'insert into blah(a, b) select "1", 1', 'transaction': True}],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'extra': 'even more',
            'strategy': 'check',
            'check_cols': ['a', 'b'],
        }
        cfg = self.ContractType(
            column_types={'a': 'text'},
            materialized='table',
            post_hook=[Hook(sql='insert into blah(a, b) select "1", 1')],
            strategy=SnapshotStrategy.Check,
            check_cols=['a', 'b'],
            target_database='some_snapshot_db',
            target_schema='some_snapshot_schema',
            unique_key='id',
        )
        cfg._extra['extra'] = 'even more'

        self.assert_symmetric(cfg, cfg_dict)

    def test_invalid_wrong_strategy(self):
        wrong_strategy = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'strategy': 'timestamp',
            'check_cols': 'all',
        }
        self.assert_fails_validation(wrong_strategy)

    def test_invalid_missing_check_cols(self):
        wrong_fields = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'strategy': 'check',
            'updated_at': 'last_update'
        }
        self.assert_fails_validation(wrong_fields)

    def test_invalid_check_value(self):
        invalid_check_type = {
            'column_types': {},
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'quoting': {},
            'tags': [],
            'vars': {},
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
            'unique_key': 'id',
            'strategy': 'timestamp',
            'check_cols': 'some',
        }
        self.assert_fails_validation(invalid_check_type)


class TestParsedSnapshotNode(ContractTestCase):
    ContractType = ParsedSnapshotNode

    def test_timestamp_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Snapshot),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'target_database': 'some_snapshot_db',
                'target_schema': 'some_snapshot_schema',
                'unique_key': 'id',
                'strategy': 'timestamp',
                'updated_at': 'last_update',
            },
            'docrefs': [],
            'columns': {},
        }

        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Snapshot,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=TimestampSnapshotConfig(
                strategy=SnapshotStrategy.Timestamp,
                unique_key='id',
                updated_at='last_update',
                target_database='some_snapshot_db',
                target_schema='some_snapshot_schema',
            ),
        )

        cfg = NodeConfig()
        cfg._extra.update({
            'unique_key': 'id',
            'strategy': 'timestamp',
            'updated_at': 'last_update',
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
        })

        inter = IntermediateSnapshotNode(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Snapshot,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=cfg,
        )
        self.assert_symmetric(node, node_dict)
        self.assert_symmetric(inter, node_dict, cls=IntermediateSnapshotNode)
        self.assertEqual(
            self.ContractType.from_dict(inter.to_dict()),
            node
        )
        self.assertTrue(node.is_refable)
        self.assertFalse(node.is_ephemeral)
        pickle.loads(pickle.dumps(node))

    def test_check_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Snapshot),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'target_database': 'some_snapshot_db',
                'target_schema': 'some_snapshot_schema',
                'unique_key': 'id',
                'strategy': 'check',
                'check_cols': 'all',
            },
            'docrefs': [],
            'columns': {},
        }

        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Snapshot,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=CheckSnapshotConfig(
                strategy=SnapshotStrategy.Check,
                unique_key='id',
                check_cols=All.All,
                target_database='some_snapshot_db',
                target_schema='some_snapshot_schema',
            ),
        )
        cfg = NodeConfig()
        cfg._extra.update({
            'unique_key': 'id',
            'strategy': 'check',
            'check_cols': 'all',
            'target_database': 'some_snapshot_db',
            'target_schema': 'some_snapshot_schema',
        })

        inter = IntermediateSnapshotNode(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Snapshot,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=cfg,
        )
        self.assert_symmetric(node, node_dict)
        self.assert_symmetric(inter, node_dict, cls=IntermediateSnapshotNode)
        self.assertEqual(
            self.ContractType.from_dict(inter.to_dict()),
            node
        )
        self.assertTrue(node.is_refable)
        self.assertFalse(node.is_ephemeral)

    def test_invalid_bad_resource_type(self):
        bad_resource_type = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'target_database': 'some_snapshot_db',
                'target_schema': 'some_snapshot_schema',
                'unique_key': 'id',
                'strategy': 'timestamp',
                'updated_at': 'last_update',
            },
            'docrefs': [],
            'columns': {},
        }
        self.assert_fails_validation(bad_resource_type)


class TestParsedNodePatch(ContractTestCase):
    ContractType = ParsedNodePatch

    def test_empty(self):
        dct = {
            'name': 'foo',
            'description': 'The foo model',
            'original_file_path': '/path/to/schema.yml',
            'columns': {},
            'docrefs': [],
        }
        patch = ParsedNodePatch(
            name='foo',
            description='The foo model',
            original_file_path='/path/to/schema.yml',
            columns={},
            docrefs=[],
        )
        self.assert_symmetric(patch, dct)

    def test_populated(self):
        dct = {
            'name': 'foo',
            'description': 'The foo model',
            'original_file_path': '/path/to/schema.yml',
            'columns': {'a': {'name': 'a', 'description': 'a text field'}},
            'docrefs': [
                {
                    'documentation_name': 'foo',
                    'documentation_package': 'test',
                }
            ],
        }
        patch = ParsedNodePatch(
            name='foo',
            description='The foo model',
            original_file_path='/path/to/schema.yml',
            columns={'a': ColumnInfo(name='a', description='a text field')},
            docrefs=[
                Docref(documentation_name='foo', documentation_package='test'),
            ],
        )
        self.assert_symmetric(patch, dct)
        pickle.loads(pickle.dumps(patch))


class TestParsedMacro(ContractTestCase):
    ContractType = ParsedMacro

    def test_ok(self):
        macro_dict = {
            'name': 'foo',
            'path': '/root/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': '{% macro foo() %}select 1 as id{% endmacro %}',
            'root_path': '/root/',
            'resource_type': 'macro',
            'unique_id': 'macro.test.foo',
            'tags': [],
            'depends_on': {'macros': []}
        }
        macro = ParsedMacro(
            name='foo',
            path='/root/path.sql',
            original_file_path='/root/path.sql',
            package_name='test',
            raw_sql='{% macro foo() %}select 1 as id{% endmacro %}',
            root_path='/root/',
            resource_type=NodeType.Macro,
            unique_id='macro.test.foo',
            tags=[],
            depends_on=MacroDependsOn()
        )
        self.assert_symmetric(macro, macro_dict)
        self.assertEqual(macro.local_vars(), {})
        pickle.loads(pickle.dumps(macro))

    def test_invalid_missing_unique_id(self):
        bad_missing_uid = {
            'name': 'foo',
            'path': '/root/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': '{% macro foo() %}select 1 as id{% endmacro %}',
            'root_path': '/root/',
            'resource_type': 'macro',
            'tags': [],
            'depends_on': {'macros': []}
        }
        self.assert_fails_validation(bad_missing_uid)

    def test_invalid_extra_field(self):
        bad_extra_field = {
            'name': 'foo',
            'path': '/root/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': '{% macro foo() %}select 1 as id{% endmacro %}',
            'root_path': '/root/',
            'resource_type': 'macro',
            'unique_id': 'macro.test.foo',
            'tags': [],
            'depends_on': {'macros': []},
            'extra': 'too many fields'
        }
        self.assert_fails_validation(bad_extra_field)


class TestParsedDocumentation(ContractTestCase):
    ContractType = ParsedDocumentation

    def test_ok(self):
        doc_dict = {
            'block_contents': 'some doc contents',
            'file_contents': '{% doc foo %}some doc contents{% enddoc %}',
            'name': 'foo',
            'original_file_path': '/root/docs/doc.md',
            'package_name': 'test',
            'path': '/root/docs',
            'root_path': '/root',
            'unique_id': 'test.foo',
        }
        doc = self.ContractType(
            package_name='test',
            root_path='/root',
            path='/root/docs',
            original_file_path='/root/docs/doc.md',
            file_contents='{% doc foo %}some doc contents{% enddoc %}',
            name='foo',
            unique_id='test.foo',
            block_contents='some doc contents'
        )
        self.assert_symmetric(doc, doc_dict)
        pickle.loads(pickle.dumps(doc))

    def test_invalid_missing(self):
        bad_missing_contents = {
            # 'block_contents': 'some doc contents',
            'file_contents': '{% doc foo %}some doc contents{% enddoc %}',
            'name': 'foo',
            'original_file_path': '/root/docs/doc.md',
            'package_name': 'test',
            'path': '/root/docs',
            'root_path': '/root',
            'unique_id': 'test.foo',
        }
        self.assert_fails_validation(bad_missing_contents)

    def test_invalid_extra(self):
        bad_extra_field = {
            'block_contents': 'some doc contents',
            'file_contents': '{% doc foo %}some doc contents{% enddoc %}',
            'name': 'foo',
            'original_file_path': '/root/docs/doc.md',
            'package_name': 'test',
            'path': '/root/docs',
            'root_path': '/root',
            'unique_id': 'test.foo',

            'extra': 'more',
        }
        self.assert_fails_validation(bad_extra_field)


class TestParsedSourceDefinition(ContractTestCase):
    ContractType = ParsedSourceDefinition

    def test_basic(self):
        source_def_dict = {
            'package_name': 'test',
            'root_path': '/root',
            'path': '/root/models/sources.yml',
            'original_file_path': '/root/models/sources.yml',
            'database': 'some_db',
            'schema': 'some_schema',
            'fqn': ['test', 'source', 'my_source', 'my_source_table'],
            'source_name': 'my_source',
            'name': 'my_source_table',
            'source_description': 'my source description',
            'loader': 'stitch',
            'identifier': 'my_source_table',
            'resource_type': str(NodeType.Source),
            'description': '',
            'docrefs': [],
            'columns': {},
            'quoting': {},
            'unique_id': 'test.source.my_source.my_source_table',
        }
        source_def = self.ContractType(
            columns={},
            docrefs=[],
            database='some_db',
            description='',
            fqn=['test', 'source', 'my_source', 'my_source_table'],
            identifier='my_source_table',
            loader='stitch',
            name='my_source_table',
            original_file_path='/root/models/sources.yml',
            package_name='test',
            path='/root/models/sources.yml',
            quoting=Quoting(),
            resource_type=NodeType.Source,
            root_path='/root',
            schema='some_schema',
            source_description='my source description',
            source_name='my_source',
            unique_id='test.source.my_source.my_source_table',
        )
        self.assert_symmetric(source_def, source_def_dict)
        minimum = {
            'package_name': 'test',
            'root_path': '/root',
            'path': '/root/models/sources.yml',
            'original_file_path': '/root/models/sources.yml',
            'database': 'some_db',
            'schema': 'some_schema',
            'fqn': ['test', 'source', 'my_source', 'my_source_table'],
            'source_name': 'my_source',
            'name': 'my_source_table',
            'source_description': 'my source description',
            'loader': 'stitch',
            'identifier': 'my_source_table',
            'resource_type': str(NodeType.Source),
            'unique_id': 'test.source.my_source.my_source_table',
        }
        self.assert_from_dict(source_def, minimum)
        pickle.loads(pickle.dumps(source_def))

    def test_invalid_missing(self):
        bad_missing_name = {
            'package_name': 'test',
            'root_path': '/root',
            'path': '/root/models/sources.yml',
            'original_file_path': '/root/models/sources.yml',
            'database': 'some_db',
            'schema': 'some_schema',
            'fqn': ['test', 'source', 'my_source', 'my_source_table'],
            'source_name': 'my_source',
            # 'name': 'my_source_table',
            'source_description': 'my source description',
            'loader': 'stitch',
            'identifier': 'my_source_table',
            'resource_type': str(NodeType.Source),
            'unique_id': 'test.source.my_source.my_source_table',
        }
        self.assert_fails_validation(bad_missing_name)

    def test_invalid_bad_resource_type(self):
        bad_resource_type = {
            'package_name': 'test',
            'root_path': '/root',
            'path': '/root/models/sources.yml',
            'original_file_path': '/root/models/sources.yml',
            'database': 'some_db',
            'schema': 'some_schema',
            'fqn': ['test', 'source', 'my_source', 'my_source_table'],
            'source_name': 'my_source',
            'name': 'my_source_table',
            'source_description': 'my source description',
            'loader': 'stitch',
            'identifier': 'my_source_table',
            'resource_type': str(NodeType.Model),
            'unique_id': 'test.source.my_source.my_source_table',
        }
        self.assert_fails_validation(bad_resource_type)
